/* Executor to simulate the stack machine used in ASU Red Dragon Book.
 *
 * 2.10.89 Kenneth Oksanen
 *   First Version
 *
 * 15.10.89 Kenneth Oksanen
 *   Bugfix in realloc, had forgotten the "sizeof( )".
 */


#include <stdio.h>
#include "defs.h"
#include "sym.h"
#include "stack.h"


static YYSTYPE yylval; /* Extra information "returned" from lexer. */

#include "lex.yy.c" /* C generated by lex from lex.l */


/* A raw hack -style parser to read the source code lexed by yylex.
 * Could be done quite easily using e.g. state machines or yacc.
 *
 * The memory reserved by program grows when needed as the parse conti-
 * nues.
 */

instruction_t* load(void)
{
  int size = 1000, cur = 0, tmp;
  /* 1000 is the initial minimal reserved for the program */
  instruction_t *program;

  program = (instruction_t *) malloc(size * sizeof(instruction_t));

  for (;;) {
    switch (program[cur].opcode = yylex()) {
    case END_OF_PROGRAM:    /* Parse over, quit this function */
      program[cur].opcode = HALT;
        /* This in order to prevent the program running out of limits. */
      return program;
    case PLUS:
    case MINUS:
    case MULT:
    case DIV:
    case MOD:
    case POP:
    case HALT:
    case SETQ:
    case COPY:
    case PRINT:
      break;
    case PUSH:
      if (yylex() == NUMERICAL_CONSTANT)
	program[cur].arg.imm_value = yylval.number;
      else {
	fprintf(stderr, "Number expected near line %d.\n", yylineno);
	exit(1);
      }
      break;
    case LABEL: 
      /* LABEL will NOT increment pc, it only manipulates the symbol-tree.
       */
      if (yylex() == SYMBOL_CONSTANT)
	yylval.symbol->data.address = cur;
      else {
	fprintf(stderr, "Identifier expected near line %d.\n", yylineno);
	exit(1);
      } 
      break;
    case RVALUE:
    case LVALUE:
    case GOTO:
    case GOFALSE:
    case GOTRUE:
      if (yylex() == SYMBOL_CONSTANT) 
	program[cur].arg.symbol = yylval.symbol;
      else {
	fprintf(stderr, "Identifier expected near line %d.\n");
	exit(1);
      }
      break;
    default:
      fprintf(stderr, "Syntax error near line %d.\n", yylineno);
      exit(1);
    } /* switch */
    
    if (program[cur].opcode != LABEL) /* Unless this opcode increased
				       * the program enough, realloc
				       * more memory.
				       */
      if (++cur >= size) { 
	size *= 2;
	program = (instruction_t *) realloc(program, 
					    size * sizeof(instruction_t)); 
	if (program == NULL) {
	  fprintf(stderr, "Memory exhausted.\n");
	  exit(1);
	}
      }
  } /* for */
}


/* Execute the program this module is given a pointer to.
 */

void execute(instruction_t *program)
{
  int pc = 0;
  stacktype_t tmp;

  for (;;) {
#ifdef DEBUG
    fprintf(stderr, "COMMAND: %d.\n", program[pc].opcode);
#ifdef DEEP_DEBUG
    print_syms();
#endif
#endif
    switch(program[pc].opcode) {
    case PLUS:
      push(pop() + pop());
      pc++;
      break;
    case MINUS:
      tmp = pop();
      push(- tmp + pop()); /* This is how subtraction is defined ... */
      pc++;
      break;
    case MULT:
      push(pop() * pop());
      pc++;
      break;
    case DIV:
      tmp = pop();
      push(pop() / tmp);
      pc++;
      break;
    case MOD:
      tmp = pop();
      push(pop() % tmp);
      pc++;
      break;
    case POP:
      pop();
      pc++;
      break;
    case PRINT:
      emit();
      pc++;
      break;
    case COPY:
      dup();
      pc++;
      break;
    case HALT:
      stack_dump();
      return;
      break;
    case PUSH:
      push(program[pc].arg.imm_value);
      pc++;
      break;
    case RVALUE:
      push(program[pc].arg.symbol->data.value);
      pc++;
      break;
    case LVALUE:
      push_sym(program[pc].arg.symbol);
      pc++;
      break;
    case SETQ:
      tmp = pop();
      pop_sym()->data.value = tmp;
      pc++;
      break;
    case GOTO:
      pc = program[pc].arg.symbol->data.address;
      if (pc == -1) {
	fprintf(stderr, "Undefined label.\n");
	exit(1);
      }
      break;
    case GOFALSE:
      if (pop() == 0) {
	pc = program[pc].arg.symbol->data.address;
	if (pc == -1) {
	  fprintf(stderr, "Undefined label.\n");
	  exit(1);
	}
      } else
	pc++;
      break;
    case GOTRUE:
      if (pop() != 0) {
	pc = program[pc].arg.symbol->data.address;
	if (pc == -1) {
	  fprintf(stderr, "Undefined label.\n");
	  exit(1);
	}
      } else
	pc++;
      break;
    default:
      fprintf(stderr, "Unknown opcode\n");
      exit(1);
    }
  }
}


/* DISCUSSION : I have not attempted to make this system totally robust.
 *   It does some things toward it, but one could e.g. rvalue and lvalue
 *   a label... However I have tried to make it easy to continue with
 *   this program.
 */


main(int argc, const char **const argv)
{
  if (argc > 2) {
    fprintf(stderr, "Usage: %s [filename]\n", argv[0]);
    exit(1);
  }

  if (argc == 2) {
    if ((yyin = fopen(argv[1], "r")) == NULL) {
      fprintf(stderr, "Couldn't open file \"%s\".\n", argv[1]);
      exit(1);
    }
  }

  execute(load());
  return 0;
}


